"""Vorze script module."""
from abc import abstractmethod
from dataclasses import dataclass
from typing import Any
from typing import ClassVar
from typing import Iterable
from typing import Tuple
from typing import Type
from typing import TypeVar

from buttplug.core import RotateCmd
from buttplug.core import RotateSubcommand

from .base import BaseCommand

# from buttplug.core import LinearCmd
# from buttplug.core import LinearSubcommand
# from buttplug.core import SpeedSubcommand
# from buttplug.core import VibrateCmd


_T = TypeVar("_T", bound="BaseVorzeCommand")


class BaseVorzeCommand(BaseCommand):
    """Base Vorze script command.

    Supports native serialization to Vorze CSV and LPEG VCSX.
    """

    @abstractmethod
    def to_csv(self) -> Iterable[Any]:
        """Return Vorze CSV row data for this command."""

    @classmethod
    @abstractmethod
    def from_csv(cls: Type[_T], row: Iterable[Any]) -> _T:
        """Construct command from a Vorze CSV row.

        Arguments:
            row: CSV row data.
        """

    @abstractmethod
    def to_vcsx(self) -> bytes:
        """Return LPEG VCSX data for this command."""

    @classmethod
    @abstractmethod
    def from_vcsx(cls: Type[_T], data: bytes) -> _T:
        """Construct command from LPEG VCSX data.

        Arguments:
            data: VCSX data.
        """

    @classmethod
    @abstractmethod
    def vcsx_size(cls) -> int:
        """Return size of VCSX command data in bytes."""


@dataclass
class VorzeRotateCommand(BaseVorzeCommand):
    """Vorze rotation command.

    Attributes:
        speed: Rotation speed with a range of [0-100].
        clockwise: Rotation direction.
    """

    SPEED_DIVISOR: ClassVar[int] = 100

    speed: int
    clockwise: bool

    def to_buttplug(self, device_index: int = 0) -> RotateCmd:
        """Return this command as a Buttplug message.

        Arguments:
            device_index: Buttplug device index.

        Returns:
            Buttplug message.
        """
        cmd = RotateSubcommand(0, self.speed / self.SPEED_DIVISOR, self.clockwise)
        return RotateCmd(device_index, [cmd])

    @classmethod
    def from_buttplug(cls, cmd: RotateCmd) -> "VorzeRotateCommand":
        """Construct command from a Buttplug message."""
        rotation = cmd.rotations[0]
        return VorzeRotateCommand(
            rotation.speed * cls.SPEED_DIVISOR, rotation.clockwise
        )

    def to_csv(self) -> Tuple[int, int]:
        """Return Vorze CSV row data for this command."""
        return 0 if self.clockwise else 1, self.speed

    @classmethod
    def from_csv(
        cls: Type["VorzeRotateCommand"], row: Iterable[Any]
    ) -> "VorzeRotateCommand":
        """Construct command from a Vorze CSV row.

        Arguments:
            row: CSV row data.

        Returns:
            Rotation command.
        """
        direction, speed = row
        return cls(int(speed), int(direction) == 0)

    @classmethod
    def vcsx_size(cls) -> int:
        """Return size of VCSX command data in bytes."""
        return 1

    def to_vcsx(self) -> bytes:
        """Return LPEG VCSX data for this command."""
        cmd = (self.speed & 0x7F) | (0 if self.clockwise else 0x80)
        return bytes([cmd])

    @classmethod
    def from_vcsx(cls: Type["VorzeRotateCommand"], data: bytes) -> "VorzeRotateCommand":
        """Construct command from LPEG VCSX data.

        Arguments:
            data: VCSX data.

        Returns:
            Rotation command.

        Raises:
            ValueError: Invalid VCSX data.
        """
        if not data:
            raise ValueError("Invalid VCSX data")
        cmd = data[0]
        speed = cmd & 0x7F
        clockwise = cmd & 0x80 == 0
        return cls(speed, clockwise)



"""Generic command module."""
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass

from buttplug.core import ButtplugMessage
from buttplug.core import LinearCmd
from buttplug.core import LinearSubcommand
from buttplug.core import RotateCmd
from buttplug.core import RotateSubcommand
from buttplug.core import SpeedSubcommand
from buttplug.core import VibrateCmd


class BaseCommand(ABC):
    """Base (native) script command."""

    @abstractmethod
    def to_buttplug(self, device_index: int = 0) -> ButtplugMessage:
        """Return this command as a Buttplug message.

        Arguments:
            device_index: Buttplug device index.
        """

    @classmethod
    @abstractmethod
    def from_buttplug(cls, msg: ButtplugMessage) -> "BaseCommand":
        """Construct command from a Buttplug message."""


@dataclass
class VibrateCommand(BaseCommand):
    """Generic vibration command.

    Attributes
        speed: Vibration speed with a range of [0.0-1.0].
    """

    speed: float

    def to_buttplug(self, device_index: int = 0) -> VibrateCmd:
        """Return this command as a Buttplug message.

        Arguments:
            device_index: Buttplug device index.

        Returns:
            Buttplug message.
        """
        return VibrateCmd(device_index, [SpeedSubcommand(0, self.speed)])

    @classmethod
    def from_buttplug(cls, msg: VibrateCmd) -> "VibrateCommand":
        """Construct command from a Buttplug message."""
        cmd = msg.speeds[0]
        return cls(cmd.speed)


@dataclass
class LinearCommand(BaseCommand):
    """Generic linear movement command.

    Attributes:
        duration: Movement time in milliseconds.
        position: Target position with a range of [0.0-1.0].
    """

    duration: int
    position: float

    def to_buttplug(self, device_index: int = 0) -> LinearCmd:
        """Return this command as a Buttplug message.

        Arguments:
            device_index: Buttplug device index.

        Returns:
            Buttplug message.
        """
        return LinearCmd(
            device_index, [LinearSubcommand(0, self.duration, self.position)]
        )

    @classmethod
    def from_buttplug(cls, msg: LinearCmd) -> "LinearCommand":
        """Construct command from a Buttplug message."""
        cmd = msg.vectors[0]
        return cls(cmd.duration, cmd.position)


@dataclass
class RotateCommand(BaseCommand):
    """Generic rotation command.

    Attributes:
        speed: Rotation speed with a range of [0.0-1.0].
        clockwise: Direction of rotation.
    """

    speed: float
    clockwise: bool

    def to_buttplug(self, device_index: int = 0) -> RotateCmd:
        """Return this command as a Buttplug message.

        Arguments:
            device_index: Buttplug device index.

        Returns:
            Buttplug message.
        """
        return RotateCmd(
            device_index, [RotateSubcommand(0, self.speed, self.clockwise)]
        )

    @classmethod
    def from_buttplug(cls, msg: RotateCmd) -> "RotateCommand":
        """Construct command from a Buttplug message."""
        cmd = msg.rotations[0]
        return cls(cmd.speed, cmd.clockwise)



"""Vorze CSV script module."""
import csv
import io
from abc import abstractmethod
from dataclasses import dataclass
from typing import BinaryIO
from typing import Type
from typing import TypeVar

from typing_extensions import TypeAlias

from ..command.vorze import BaseVorzeCommand
from ..command.vorze import VorzeRotateCommand
from ..exceptions import ParseError
from .base import ScriptCommand
from .base import SerializableScript


_T = TypeVar("_T", bound=BaseVorzeCommand)


@dataclass
class VorzeScriptCommand(ScriptCommand[_T]):
    """Timed Vorze script command."""


_SC: TypeAlias = VorzeScriptCommand


class VorzeScript(SerializableScript[_T]):
    """Generic Vorze CSV script.

    Note:
        Vorze CSVs time offsets are integers in tenths of a second. Internally
        we store offsets in milliseconds so loss of resolution will occur when
        converting to CSV.
    """

    OFFSET_DENOM = 100

    @classmethod
    @abstractmethod
    def _command_cls(cls) -> Type[_T]:
        """Return command class for this script."""

    def dump(self, fp: BinaryIO) -> None:
        """Serialize script to file.

        Arguments:
            fp: A file-like object opened for writing.
        """
        text_fp = io.TextIOWrapper(fp, newline="")
        try:
            writer = csv.writer(text_fp)
            writer.writerows(
                (self.offset_from_ms(cmd.offset),) + tuple(cmd.cmd.to_csv())
                for cmd in self.commands
            )
        finally:
            text_fp.detach()

    @classmethod
    def load(cls, fp: BinaryIO) -> "VorzeScript[_T]":
        """Deserialize script from file.

        Arguments:
            fp: A file-like object opened for reading.

        Returns:
            Loaded command script.

        Raises:
            ParseError: A CSV parsing error occured.
        """
        try:
            reader = csv.reader(io.TextIOWrapper(fp, newline=""))
            return cls(
                VorzeScriptCommand(
                    cls.offset_to_ms(int(row[0])),
                    cls._command_cls().from_csv(row[1:]),
                )
                for row in reader
            )
        except (csv.Error, ValueError) as e:
            raise ParseError("Failed to parse file as Vorze CSV.") from e

    @classmethod
    def offset_from_ms(cls, offset_ms: int) -> int:
        """Convert millisecond time offset to Vorze CSV offsets."""
        return round(offset_ms / cls.OFFSET_DENOM)

    @classmethod
    def offset_to_ms(cls, offset_csv: int) -> int:
        """Convert Vorze CSV time offsets to millisecond offsets."""
        return offset_csv * cls.OFFSET_DENOM


class VorzeRotateScript(VorzeScript[VorzeRotateCommand]):
    """Vorze rotation commands script.

    Note:
        Shared script format for both Cyclone and UFO devices.
    """

    @classmethod
    def _command_cls(cls) -> Type[VorzeRotateCommand]:
        """Return command class for this script."""
        return VorzeRotateCommand



"""Device command module."""
from .base import BaseCommand
from .base import LinearCommand
from .base import RotateCommand
from .base import VibrateCommand
from .vorze import VorzeRotateCommand


__all__ = [
    "BaseCommand",
    "LinearCommand",
    "RotateCommand",
    "VibrateCommand",
    "VorzeRotateCommand",
]
